/***************************************************************************
                          main.c  -  Demo program for 617
                             -------------------
    begin                : Thu Feb 10 2005
    copyright            : (C) 2005 by Sensoray
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/poll.h>
#include <time.h>

#include "sx17.h"

//  The following 2 struct's are required for saving
//  uncompressed image in Windows bitmap format
#pragma pack(1)
typedef struct tagBITMAPFILEHEADER {
  WORD    bfType;
  DWORD   bfSize;
  WORD    bfReserved1;
  WORD    bfReserved2;
  DWORD   bfOffBits;
} BITMAPFILEHEADER, *PBITMAPFILEHEADER;

typedef struct tagBITMAPINFOHEADER{
  DWORD  biSize;
  LONG   biWidth;
  LONG   biHeight;
  WORD   biPlanes;
  WORD   biBitCount;
  DWORD  biCompression;
  DWORD  biSizeImage;
  LONG   biXPelsPerMeter;
  LONG   biYPelsPerMeter;
  DWORD  biClrUsed;
  DWORD  biClrImportant;
} BITMAPINFOHEADER, *PBITMAPINFOHEADER;
#pragma pack()

void SaveBMP (char *fname, void *bmpdata, MODE *mode);
void SaveJPG (char *fname, void *jpgdata, unsigned long jpgsize);
void PrintBfr (BUFFER *bfr);
void PrintCmd (COMMAND *cmd);
int next_chan (void);

int main(int argc, char *argv[])
{
    int             fd;     //file descriptor
    char            *device_name = "/dev/sx17a";
    unsigned long   ecode;
    int             i;
    int             rc;
    void            *map;
    unsigned long   mapsize;
    OFFSET          offsets[SYS_BUFS];
    MODE            mode = DEF_MODE;
    COMMAND         cmd;
    BUFFER          bfr;
    struct pollfd   ufds;
    void            *bmpdata, *jpgdata;
    char            fname[80];
    char            caption[65];    //buffer for caption manipulations
    struct timeval  tv;             //used for timestamp
    struct tm       *ptm;           //used for timestamp

    //Open sx17 device
    if ((fd = open (device_name, O_RDONLY | O_NONBLOCK )) <= 0) {
        printf ("OPEN ERROR\n");
    } else {
        //set up poll struct
        ufds.fd = fd;               //our device
        ufds.events = POLLIN;       //the only poll flag supported
        //get the size of driver memory that has to be mapped
        mapsize = ioctl (fd, SX17_IOC_GETMAPSIZE);
        //map driver memory into application's space
        if ((map = mmap(0, mapsize, PROT_READ, MAP_SHARED, fd, 0)) == (char *) -1){
            printf ("mmap error\n");
            return EXIT_SUCCESS;
        }
        //get offsets to buffers
        ioctl (fd, SX17_IOC_GETOFFSETS, offsets);
        //set operation mode (using default)
        mode.bmpsize = BMP_SIZE_SML;
        mode.jpgsize = JPG_SIZE_FULL;
        ecode = ioctl (fd, SX17_IOC_SETMODE, &mode);
        if (ecode) {
            printf ("Setmode error %ld\n", ecode);
            return EXIT_SUCCESS;
        }

        //Set up COMMAND struct
	    cmd.jpegon =        JPEG_ON;
        cmd.bmpon =         BMP_ON;
        cmd.compfactor =    0x100;
        cmd.bright =        0x80;
        cmd.contrast =      0x44;
        cmd.sat =           0x40;
        cmd.hue =           0;
    	cmd.captionmode =   CAP_ON | CAP_WHT;
        cmd.vxp =			0x01020304;
        cmd.offset =        0;  //board 1; cameras 1-16
        cmd.pCaption =      "";

        //Queue 2 commands to make sure the queue does not run out
        cmd.xpchan =        next_chan ();
        ioctl (fd, SX17_IOC_QUEUECMD, &cmd);
        cmd.xpchan =        next_chan ();
        ioctl (fd, SX17_IOC_QUEUECMD, &cmd);

		//Read data, queue next command
        for (i = 0; i < 30; i++) {
	        //Poll for data
            rc = poll (&ufds, 1, 2000);				//2 sec timeout
            if (rc <= 0) {
                if (rc == 0)
                    printf ("Poll timed out\n");
                else
                    printf ("Poll error \n");
                return EXIT_SUCCESS;
            }

            rc = ioctl (fd, SX17_IOC_GETBUFR, &bfr);	//returns non-zero if OK
            if (rc) {
				if (bfr.valid) {						//check if valid data
                    if (bfr.videopresent) {
                        bmpdata = (void *) ((long) map + offsets[bfr.bi].obmp);		//pointer to bitmap data
                        jpgdata = (void *) ((long) map + offsets[bfr.bi].ojpg);		//pointer to jpeg data
                        sprintf (fname, "%02d.bmp", bfr.xpchan);						//generate file name
                        SaveBMP (fname, bmpdata, &mode);									//save bitmap
                        if (bfr.jpegsize) {											//if valid jpeg data
                            sprintf (fname, "%02d.jpg", bfr.xpchan);
                            SaveJPG (fname, jpgdata, bfr.jpegsize);					//save jpeg file
                        }
                    } else {
                        printf ("No video on chan %d\n", bfr.xpchan);
                    }
				} else {
					printf ("Invalid buffer xpc %d ch %d Inc %d Zerr %d\n", bfr.xpchan, bfr.reserved[0], bfr.incomplete, bfr.zerr);
				}
	            ioctl (fd, SX17_IOC_RELEASEBUFR, (void *) bfr.bi);				//we're done with this buffer
            } else {
                printf ("No buffer\n");
			}

			//Queue next cmnd
            cmd.xpchan = next_chan ();
            gettimeofday (&tv, NULL);
            ptm = localtime (&tv.tv_sec);                                       //convert into readable format
            strftime (caption, 64, "Camera ^c^n%Y-%m-%d^n%H:%M:%S", ptm);       //prepare caption string
            cmd.pCaption = caption;
            ioctl (fd, SX17_IOC_QUEUECMD, &cmd);
        }

        //close driver
        close (fd);
    }
    return EXIT_SUCCESS;
}

////////////////////////////////////////////////////////////////
// Saves color bitmap to a file
////////////////////////////////////////////////////////////////
void SaveBMP (char *fname, void *bmpdata, MODE *mode)
{
    BITMAPFILEHEADER            bfh;
    BITMAPINFOHEADER            bih;
    const long                  SIZE_BFH = sizeof (BITMAPFILEHEADER);
    const long                  SIZE_BIH = sizeof (BITMAPINFOHEADER);
    long                        bmpsize;
    FILE                        *file;

    memset (&bfh, 0, sizeof (BITMAPFILEHEADER));
    memset (&bih, 0, sizeof (BITMAPINFOHEADER));

    bih.biSize = sizeof (BITMAPINFOHEADER);
	switch (mode->bmpsize) {
    	case BMP_SIZE_HUG:
		    bih.biWidth = 640;
    		bih.biHeight = 240;
            break;
    	case BMP_SIZE_LRG:
		    bih.biWidth = 256;
    		bih.biHeight = 192;
            break;
    	case BMP_SIZE_SML:
        default:
		    bih.biWidth = 128;
    		bih.biHeight = 96;
            break;
	}

    bih.biPlanes = 1;
    bih.biCompression = 0;
    bih.biSizeImage = 0;
    bih.biXPelsPerMeter = 0;
    bih.biYPelsPerMeter = 0;
    bih.biBitCount = 24;
    bih.biClrUsed = 0;
    bih.biClrImportant = 0;

    bmpsize = bih.biWidth * bih.biHeight * bih.biBitCount / 8;

    bfh.bfType = * (WORD *) "BM";
    bfh.bfOffBits = SIZE_BFH + SIZE_BIH;
    bfh.bfSize =  bfh.bfOffBits + bmpsize;

    file = fopen (fname, "w");
    if (file) {
        fwrite (&bfh, 1, SIZE_BFH, file);
        fwrite (&bih, 1, SIZE_BIH, file);
        fwrite (bmpdata, 1, bmpsize, file);
        fclose (file);
		printf ("File saved %s, %d bytes\n", fname, bmpsize);
    }
    return;
}

//Saves JPEG file
void SaveJPG (char *fname, void *jpgdata, unsigned long jpgsize)
{
    FILE		*file;

    file = fopen (fname, "w");
    if (file) {
        fwrite (jpgdata, 1, jpgsize, file);
        fclose (file);
		printf ("File saved %s, %d bytes\n", fname, jpgsize);
    }
    return;
}


//Helper functions
void PrintBfr (BUFFER *bfr)
{
    printf ("BUFFER:\n");
	printf ("bi            %d\n", bfr->bi);
    printf ("valid         %d\n", bfr->valid);
	printf ("jpegsize      %ld\n", bfr->jpegsize);
	printf ("xpchan        %d\n", bfr->xpchan);
    printf ("vxp           0x%08X\n", bfr->vxp);
	printf ("videopresent  0x%08X\n", bfr->videopresent);
    printf ("dioin         0x%08lX\n", bfr->dioin);
	printf ("compfactor    0x%08X\n", bfr->compfactor);
    printf ("bright        0x%08X\n", bfr->bright);
    printf ("contrast      0x%08X\n", bfr->contrast);
    printf ("sat           0x%08X\n", bfr->sat);
    printf ("hue           0x%08X\n", bfr->hue);
	printf ("incomplete    %d\n", bfr->incomplete);
	printf ("zerr          %d\n", bfr->zerr);
    printf ("\n");
}

void PrintCmd (COMMAND *cmd)
{
    printf ("COMMAND:\n");
	printf ("jpegon        %d\n", cmd->jpegon);
    printf ("bmpon         %d\n", cmd->bmpon);
	printf ("xpchan        %d\n", cmd->xpchan);
	printf ("compfactor    0x%08X\n", cmd->compfactor);
    printf ("bright        0x%08X\n", cmd->bright);
    printf ("contrast      0x%08X\n", cmd->contrast);
    printf ("sat           0x%08X\n", cmd->sat);
    printf ("hue           0x%08X\n", cmd->hue);
	printf ("captionmode   %d\n", cmd->captionmode);
	printf ("caption       %s\n", (char *) cmd->pCaption);
	printf ("offset        %d\n", cmd->offset);
    printf ("vxp           0x%08X\n", cmd->vxp);
    printf ("dioout        0x%08X\n", cmd->dioout);
    printf ("\n");
}

#define SEQUENCE_LENGTH			16
//#define SEQUENCE_LENGTH			3
int next_chan (void)
{
//This function generates next camera number (xpchan)
//based on the required polling sequence.
//The sequence and implementation of next_chan() are arbitrary.
//It is recommended to exclude channels that don't have a camera
//connected to them from the sequence.
	static int	sequence[SEQUENCE_LENGTH] = {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16};
//	static int	sequence[SEQUENCE_LENGTH] = {1, 2, 4};
    static int	index = 0;
	int			xpchan;

    xpchan = sequence[index];
	//increment index, roll over at (SEQUENCE_LENGTH-1)
    if (++index >= SEQUENCE_LENGTH)
    	index = 0;
    return xpchan;
}

